@@ -14,27 +14,34 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import logging
-from http.cookiejar import CookieJar
+import gzip
 import json as json_import
-from os.path import basename
+import logging
 import platform
-import sys
+from http.cookiejar import CookieJar
+from os.path import basename
 from typing import Dict, List, Optional, Tuple, Union
+from urllib3.util.retry import Retry
 
 import requests
+from requests.adapters import HTTPAdapter
 from requests.structures import CaseInsensitiveDict
+
 from ibm_cloud_sdk_core.authenticators import Authenticator
-from .version import __version__
-from .utils import has_bad_first_or_last_char, remove_null_values, cleanup_values, read_external_sources
-from .detailed_response import DetailedResponse
 from .api_exception import ApiException
-from .jwt_token_manager import JWTTokenManager
+from .detailed_response import DetailedResponse
+from .token_managers.token_manager import TokenManager
+from .utils import (has_bad_first_or_last_char, remove_null_values,
+                    cleanup_values, read_external_sources, strip_extra_slashes)
+from .version import __version__
 
 # Uncomment this to enable http debugging
 # import http.client as http_client
 # http_client.HTTPConnection.debuglevel = 1
 
+
+#pylint: disable=too-many-instance-attributes
+#pylint: disable=too-many-locals
 class BaseService:
     """Common functionality shared by generated service classes.
 
@@ -46,6 +53,7 @@ class BaseService:
         authenticator: Adds authentication data to service requests. Defaults to None.
         disable_ssl_verification: A flag that indicates whether verification of the server's SSL
             certificate should be disabled or not. Defaults to False.
+        enable_gzip_compression: A flag that indicates whether to enable gzip compression on request bodies
 
     Attributes:
         service_url (str): Url to the service endpoint.
@@ -55,6 +63,9 @@ class BaseService:
         default_headers (dict): A dictionary of headers to be sent with every HTTP request to the service endpoint.
         jar (http.cookiejar.CookieJar): Stores cookies received from the service.
         http_config (dict): A dictionary containing values that control the timeout, proxies, and etc of HTTP requests.
+        http_client (Session): A configurable session which can use Transport Adapters to configure retries, timeouts,
+            proxies, etc. globally for all requests.
+        enable_gzip_compression (bool): A flag that indicates whether to enable gzip compression on request bodies
     Raises:
         ValueError: If Authenticator is not provided or invalid type.
     """
@@ -65,35 +76,72 @@ class BaseService:
                             'disable_ssl_verification option of the authenticator.'
 
     def __init__(self,
+                 *,
                  service_url: str = None,
                  authenticator: Authenticator = None,
-                 disable_ssl_verification: bool = False):
-        self.service_url = service_url
+                 disable_ssl_verification: bool = False,
+                 enable_gzip_compression: bool = False) -> None:
+        self.set_service_url(service_url)
+        self.http_client = requests.Session()
         self.http_config = {}
         self.jar = CookieJar()
         self.authenticator = authenticator
         self.disable_ssl_verification = disable_ssl_verification
         self.default_headers = None
+        self.enable_gzip_compression = enable_gzip_compression
         self._set_user_agent_header(self._build_user_agent())
+        self.retry_config = None
+        self.http_adapter = HTTPAdapter()
         if not self.authenticator:
             raise ValueError('authenticator must be provided')
         if not isinstance(self.authenticator, Authenticator):
-            raise ValueError(
-                'authenticator should be of type Authenticator')
+            raise ValueError('authenticator should be of type Authenticator')
+
+    def enable_retries(self, max_retries: int = 4, retry_interval: float = 1.0) -> None:
+        """Enable automatic retries on the underlying http client used by the BaseService instance.
+
+        Args:
+          max_retries: the maximum number of retries to attempt for a failed retryable request
+          retry_interval: the default wait time (in seconds) to use for the first retry attempt.
+            In general, if a response includes the Retry-After header, that will be used for
+            the wait time associated with the retry attempt.  If the Retry-After header is not
+            present, then the wait time is based on the retry_interval and retry attempt number:
+               wait_time = retry_interval * (2 ^ (n-1)), where n is the retry attempt number
+        """
+        self.retry_config = Retry(
+            total=max_retries,
+            backoff_factor=retry_interval,
+            # List of HTTP status codes to retry on in addition to Timeout/Connection Errors
+            status_forcelist=[429, 500, 502, 503, 504],
+            # List of HTTP methods to retry on
+            # Omitting this will default to all methods except POST
+            allowed_methods=['HEAD', 'GET', 'PUT',
+                             'DELETE', 'OPTIONS', 'TRACE', 'POST']
+        )
+        self.http_adapter = HTTPAdapter(max_retries=self.retry_config)
+        self.http_client.mount('http://', self.http_adapter)
+        self.http_client.mount('https://', self.http_adapter)
+
+    def disable_retries(self):
+        """Remove retry config from http_adapter"""
+        self.retry_config = None
+        self.http_adapter = HTTPAdapter()
+        self.http_client.mount('http://', self.http_adapter)
+        self.http_client.mount('https://', self.http_adapter)
 
     @staticmethod
-    def _get_system_info():
+    def _get_system_info() -> str:
         return '{0} {1} {2}'.format(
             platform.system(),  # OS
             platform.release(),  # OS version
             platform.python_version()  # Python version
         )
 
-    def _build_user_agent(self):
+    def _build_user_agent(self) -> str:
         return '{0}-{1} {2}'.format(self.SDK_NAME, __version__,
                                     self._get_system_info())
 
-    def configure_service(self, service_name: str):
+    def configure_service(self, service_name: str) -> None:
         """Look for external configuration of a service. Set service properties.
 
         Try to get config from external sources, with the following priority:
@@ -115,13 +163,24 @@ class BaseService:
             self.set_service_url(config.get('URL'))
         if config.get('DISABLE_SSL'):
             self.set_disable_ssl_verification(
-                bool(config.get('DISABLE_SSL'))
-            )
+                config.get('DISABLE_SSL').lower() == 'true')
+        if config.get('ENABLE_GZIP'):
+            self.set_enable_gzip_compression(
+                config.get('ENABLE_GZIP').lower() == 'true')
+        if config.get('ENABLE_RETRIES'):
+            if config.get('ENABLE_RETRIES').lower() == 'true':
+                kwargs = {}
+                if config.get('MAX_RETRIES'):
+                    kwargs["max_retries"] = int(config.get('MAX_RETRIES'))
+                if config.get('RETRY_INTERVAL'):
+                    kwargs["retry_interval"] = float(
+                        config.get('RETRY_INTERVAL'))
+                self.enable_retries(**kwargs)
 
-    def _set_user_agent_header(self, user_agent_string=None):
+    def _set_user_agent_header(self, user_agent_string: str) -> None:
         self.user_agent_header = {'User-Agent': user_agent_string}
 
-    def set_http_config(self, http_config: dict):
+    def set_http_config(self, http_config: dict) -> None:
         """Sets the http config dictionary.
 
         The dictionary can contain values that control the timeout, proxies, and etc of HTTP requests.
@@ -134,13 +193,15 @@ class BaseService:
         """
         if isinstance(http_config, dict):
             self.http_config = http_config
-            if (self.authenticator and hasattr(self.authenticator, 'token_manager') and
-                    isinstance(self.authenticator.token_manager, JWTTokenManager)):
+            if (self.authenticator
+                    and hasattr(self.authenticator, 'token_manager')
+                    and isinstance(self.authenticator.token_manager,
+                                   TokenManager)):
                 self.authenticator.token_manager.http_config = http_config
         else:
             raise TypeError("http_config parameter must be a dictionary")
 
-    def set_disable_ssl_verification(self, status: bool = False):
+    def set_disable_ssl_verification(self, status: bool = False) -> None:
         """Set the flag that indicates whether verification of
         the server's SSL certificate should be disabled or not.
 
@@ -149,7 +210,7 @@ class BaseService:
         """
         self.disable_ssl_verification = status
 
-    def set_service_url(self, service_url: str):
+    def set_service_url(self, service_url: str) -> None:
         """Set the url the service will make HTTP requests too.
 
         Arguments:
@@ -165,6 +226,26 @@ class BaseService:
             )
         self.service_url = service_url
 
+    def get_http_client(self) -> requests.sessions.Session:
+        """Get the http client session currently used by the service.
+
+        Returns:
+            The http client session currently used by the service.
+        """
+        return self.http_client
+
+    def set_http_client(self, http_client: requests.sessions.Session) -> None:
+        """Set current http client session
+
+        Arguments:
+            http_client: A new requests session client
+        """
+        if isinstance(http_client, requests.sessions.Session):
+            self.http_client = http_client
+        else:
+            raise TypeError(
+                "http_client parameter must be a requests.sessions.Session")
+
     def get_authenticator(self) -> Authenticator:
         """Get the authenticator currently used by the service.
 
@@ -173,7 +254,7 @@ class BaseService:
         """
         return self.authenticator
 
-    def set_default_headers(self, headers: Dict[str, str]):
+    def set_default_headers(self, headers: Dict[str, str]) -> None:
         """Set http headers to be sent in every request.
 
         Arguments:
@@ -204,13 +285,26 @@ class BaseService:
         if self.disable_ssl_verification:
             kwargs['verify'] = False
 
+        # Check to see if the caller specified the 'stream' argument.
+        stream_response = kwargs.get('stream') or False
+
+        # Remove the keys we set manually, don't let the user to overwrite these.
+        reserved_keys = ['method', 'url', 'headers', 'params', 'cookies']
+        for key in reserved_keys:
+            if key in kwargs:
+                del kwargs[key]
+                logging.warning('"%s" has been removed from the request', key)
         try:
-            response = requests.request(**request, cookies=self.jar, **kwargs)
+            response = self.http_client.request(**request,
+                                                cookies=self.jar,
+                                                **kwargs)
 
             if 200 <= response.status_code <= 299:
                 if response.status_code == 204 or request['method'] == 'HEAD':
                     # There is no body content for a HEAD request or a 204 response
                     result = None
+                elif stream_response:
+                    result = response
                 elif not response.text:
                     result = None
                 else:
@@ -218,36 +312,36 @@ class BaseService:
                         result = response.json()
                     except:
                         result = response
-                return DetailedResponse(result, response.headers,
-                                        response.status_code)
+                return DetailedResponse(response=result,
+                                        headers=response.headers,
+                                        status_code=response.status_code)
 
-            error_message = None
-            if response.status_code == 401:
-                error_message = 'Unauthorized: Access is denied due to ' \
-                                'invalid credentials'
-            raise ApiException(
-                response.status_code, error_message, http_response=response)
+            raise ApiException(response.status_code, http_response=response)
         except requests.exceptions.SSLError:
             logging.exception(self.ERROR_MSG_DISABLE_SSL)
             raise
-        except ApiException as err:
-            logging.exception(err.message)
-            raise
-        except:
-            logging.exception('Error in service call')
-            raise
 
+    def set_enable_gzip_compression(self,
+                                    should_enable_compression: bool = False
+                                    ) -> None:
+        """Set value to enable gzip compression on request bodies"""
+        self.enable_gzip_compression = should_enable_compression
+
+    def get_enable_gzip_compression(self) -> bool:
+        """Get value for enabling gzip compression on request bodies"""
+        return self.enable_gzip_compression
 
     def prepare_request(self,
                         method: str,
                         url: str,
+                        *,
                         headers: Optional[dict] = None,
                         params: Optional[dict] = None,
                         data: Optional[Union[str, dict]] = None,
-                        files: Optional[Union[
-                            Dict[str, Tuple[str]],
-                            List[Tuple[str, Tuple[str, ...]]]
-                        ]] = None,
+                        files: Optional[Union[Dict[str, Tuple[str]],
+                                              List[Tuple[str,
+                                                         Tuple[str,
+                                                               ...]]]]] = None,
                         **kwargs) -> dict:
         """Build a dict that represents an HTTP service request.
 
@@ -274,7 +368,7 @@ class BaseService:
         # validate the service url is set
         if not self.service_url:
             raise ValueError('The service_url is required')
-        request['url'] = self.service_url + url
+        request['url'] = strip_extra_slashes(self.service_url + url)
 
         headers = remove_null_values(headers) if headers else {}
         headers = cleanup_values(headers)
@@ -289,17 +383,27 @@ class BaseService:
         params = cleanup_values(params)
         request['params'] = params
 
-        if sys.version_info >= (3, 0) and isinstance(data, str):
+        if isinstance(data, str):
             data = data.encode('utf-8')
-
-        if data and isinstance(data, dict):
+        elif isinstance(data, dict) and data:
             data = remove_null_values(data)
-            headers.update({'content-type': 'application/json'})
-            data = json_import.dumps(data)
+            if headers.get('content-type') is None:
+                headers.update({'content-type': 'application/json'})
+            data = json_import.dumps(data).encode('utf-8')
         request['data'] = data
 
         self.authenticator.authenticate(request)
 
+        # Compress the request body if applicable
+        if (self.get_enable_gzip_compression()
+                and 'content-encoding' not in headers
+                and request['data'] is not None):
+            headers['content-encoding'] = 'gzip'
+            uncompressed_data = request['data']
+            request_body = gzip.compress(uncompressed_data)
+            request['data'] = request_body
+            request['headers'] = headers
+
         # Next, we need to process the 'files' argument to try to fill in
         # any missing filenames where possible.
         # 'files' can be a dictionary (i.e { '<part-name>': (<tuple>)} )
@@ -313,7 +417,8 @@ class BaseService:
                 files = files.items()
             # Next, fill in any missing filenames from file tuples.
             for part_name, file_tuple in files:
-                if file_tuple and len(file_tuple) == 3 and file_tuple[0] is None:
+                if file_tuple and len(
+                        file_tuple) == 3 and file_tuple[0] is None:
                     file = file_tuple[1]
                     if file and hasattr(file, 'name'):
                         filename = basename(file.name)
@@ -340,7 +445,7 @@ class BaseService:
     # pylint: disable=protected-access
 
     @staticmethod
-    def _convert_model(val):
+    def _convert_model(val: str) -> None:
         if isinstance(val, str):
             val = json_import.loads(val)
         if hasattr(val, "_to_dict"):
@@ -348,11 +453,11 @@ class BaseService:
         return val
 
     @staticmethod
-    def _convert_list(val):
+    def _convert_list(val: list) -> None:
         if isinstance(val, list):
             return ",".join(val)
         return val
 
     @staticmethod
-    def _encode_path_vars(*args):
+    def _encode_path_vars(*args) -> None:
         return (requests.utils.quote(x, safe='') for x in args)
