@@ -1,6 +1,6 @@
 # coding: utf-8
 
-# Copyright 2019 IBM All Rights Reserved.
+# Copyright 2019, 2021 IBM All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,9 +19,11 @@ import json as json_import
 from os import getenv, environ, getcwd
 from os.path import isfile, join, expanduser
 from typing import List, Union
+from urllib.parse import urlparse, parse_qs
 
 import dateutil.parser as date_parser
 
+
 def has_bad_first_or_last_char(val: str) -> bool:
     """Returns true if a string starts with any of: {," ; or ends with any of: },".
 
@@ -31,7 +33,9 @@ def has_bad_first_or_last_char(val: str) -> bool:
     Returns:
         Whether or not the string starts or ends with bad characters.
     """
-    return val is not None and (val.startswith('{') or val.startswith('"') or val.endswith('}') or val.endswith('"'))
+    return val is not None and (val.startswith('{') or val.startswith('"')
+                                or val.endswith('}') or val.endswith('"'))
+
 
 def remove_null_values(dictionary: dict) -> dict:
     """Create a new dictionary without keys mapped to null values.
@@ -46,6 +50,7 @@ def remove_null_values(dictionary: dict) -> dict:
         return {k: v for (k, v) in dictionary.items() if v is not None}
     return dictionary
 
+
 def cleanup_values(dictionary: dict) -> dict:
     """Create a new dictionary with boolean values converted to strings.
 
@@ -62,12 +67,21 @@ def cleanup_values(dictionary: dict) -> dict:
         return {k: cleanup_value(v) for (k, v) in dictionary.items()}
     return dictionary
 
+
 def cleanup_value(value: any) -> any:
     """Convert a boolean value to string."""
     if isinstance(value, bool):
         return 'true' if value else 'false'
     return value
 
+
+def strip_extra_slashes(value: str) -> str:
+    """Combine multiple trailing slashes to a single slash"""
+    if value.endswith('//'):
+        return value.rstrip('/') + '/'
+    return value
+
+
 def datetime_to_string(val: datetime.datetime) -> str:
     """Convert a datetime object to string.
 
@@ -87,6 +101,7 @@ def datetime_to_string(val: datetime.datetime) -> str:
         return val.isoformat().replace('+00:00', 'Z')
     return val
 
+
 def string_to_datetime(string: str) -> datetime.datetime:
     """De-serializes string to datetime.
 
@@ -101,6 +116,43 @@ def string_to_datetime(string: str) -> datetime.datetime:
         return val
     return val.replace(tzinfo=datetime.timezone.utc)
 
+
+def string_to_datetime_list(string_list: List[str]) -> List[datetime.datetime]:
+    """De-serializes each string in a list to a datetime.
+
+    Args:
+        string_list: list of strings containing datetime in iso8601 format.
+
+    Returns:
+        the de-serialized list of strings as a list of datetime objects.
+    """
+    if not isinstance(string_list, list):
+        raise ValueError("Invalid argument type: " + str(type(string_list))
+            + ". Argument string_list must be of type List[str]")
+    datetime_list = []
+    for string_val in string_list:
+        datetime_list.append(string_to_datetime(string_val))
+    return datetime_list
+
+
+def datetime_to_string_list(datetime_list: List[datetime.datetime]) -> List[str]:
+    """Convert a list of datetime objects to a list of strings.
+
+    Args:
+        datetime_list: The list of datetime objects.
+
+    Returns:
+        list of datetimes serialized as strings in iso8601 format.
+    """
+    if not isinstance(datetime_list, list):
+        raise ValueError("Invalid argument type: " + str(type(datetime_list))
+            + ". Argument datetime_list must be of type List[datetime.datetime]")
+    string_list = []
+    for datetime_val in datetime_list:
+        string_list.append(datetime_to_string(datetime_val))
+    return string_list
+
+
 def date_to_string(val: datetime.date) -> str:
     """Convert a date object to string.
 
@@ -114,6 +166,7 @@ def date_to_string(val: datetime.date) -> str:
         return str(val)
     return val
 
+
 def string_to_date(string: str) -> datetime.date:
     """De-serializes string to date.
 
@@ -125,6 +178,32 @@ def string_to_date(string: str) -> datetime.date:
     """
     return date_parser.parse(string).date()
 
+
+def get_query_param(url_str: str, param: str) -> str:
+    """Return a query parameter value from url_str
+
+    Args:
+        url_str: the URL from which to extract the query
+            parameter value
+        param: the name of the query parameter whose value
+            should be returned
+
+    Returns:
+        the value of the `param` query parameter as a string
+
+    Raises:
+        ValueError: if errors are encountered parsing `url_str`
+    """
+    if not url_str:
+        return None
+    url = urlparse(url_str)
+    if not url.query:
+        return None
+    query = parse_qs(url.query, strict_parsing=True)
+    values = query.get(param)
+    return values[0] if values else None
+
+
 def convert_model(val: any) -> dict:
     """Convert a model object into an equivalent dict.
 
@@ -141,6 +220,7 @@ def convert_model(val: any) -> dict:
     # Consider raising a ValueError here in the next major release
     return val
 
+
 def convert_list(val: Union[str, List[str]]) -> str:
     """Convert a list of strings into comma-separated string.
 
@@ -157,6 +237,7 @@ def convert_list(val: Union[str, List[str]]) -> str:
     # Consider raising a ValueError here in the next major release
     return val
 
+
 def read_external_sources(service_name: str) -> dict:
     """Look for external configuration of a service.
 
@@ -183,6 +264,7 @@ def read_external_sources(service_name: str) -> dict:
 
     return config
 
+
 def __read_from_env_variables(service_name: str) -> dict:
     """Return a config object based on environment variables for a service.
 
@@ -197,7 +279,10 @@ def __read_from_env_variables(service_name: str) -> dict:
         _parse_key_and_update_config(config, service_name, key, value)
     return config
 
-def __read_from_credential_file(service_name: str, separator: str = '=') -> dict:
+
+def __read_from_credential_file(service_name: str,
+                                *,
+                                separator: str = '=') -> dict:
     """Return a config object based on credentials file for a service.
 
     Args:
@@ -211,17 +296,16 @@ def __read_from_credential_file(service_name: str, separator: str = '=') -> dict
     """
     default_credentials_file_name = 'ibm-credentials.env'
 
-    # File path specified by an env variable
+    # 1. ${IBM_CREDENTIALS_FILE}
     credential_file_path = getenv('IBM_CREDENTIALS_FILE')
 
-    # Current working directory
+    # 2. <current-working-directory>/ibm-credentials.env
     if credential_file_path is None:
-        file_path = join(
-            getcwd(), default_credentials_file_name)
+        file_path = join(getcwd(), default_credentials_file_name)
         if isfile(file_path):
             credential_file_path = file_path
 
-    # Home directory
+    # 3. <user-home-directory>/ibm-credentials.env
     if credential_file_path is None:
         file_path = join(expanduser('~'), default_credentials_file_name)
         if isfile(file_path):
@@ -229,20 +313,29 @@ def __read_from_credential_file(service_name: str, separator: str = '=') -> dict
 
     config = {}
     if credential_file_path is not None:
-        with open(credential_file_path, 'r') as fobj:
-            for line in fobj:
-                key_val = line.strip().split(separator)
-                if len(key_val) == 2:
-                    key = key_val[0]
-                    value = key_val[1]
-                    _parse_key_and_update_config(config, service_name, key, value)
+        try:
+            with open(credential_file_path, 'r', encoding='utf-8') as fobj:
+                for line in fobj:
+                    key_val = line.strip().split(separator, 1)
+                    if len(key_val) == 2:
+                        key = key_val[0]
+                        value = key_val[1]
+                        _parse_key_and_update_config(config, service_name, key,
+                                                     value)
+        except OSError:
+            # just absorb the exception and make sure we return an empty response
+            config = {}
+
     return config
 
-def _parse_key_and_update_config(config, service_name, key, value):
+
+def _parse_key_and_update_config(config: dict, service_name: str, key: str,
+                                 value: str) -> None:
     service_name = service_name.replace(' ', '_').replace('-', '_').upper()
     if key.startswith(service_name):
         config[key[len(service_name) + 1:]] = value
 
+
 def __read_from_vcap_services(service_name: str) -> dict:
     """Return a config object based on the vcap services environment variable.
 
@@ -258,29 +351,39 @@ def __read_from_vcap_services(service_name: str) -> dict:
         services = json_import.loads(vcap_services)
         for key in services.keys():
             for i in range(len(services[key])):
-                if vcap_service_credentials and isinstance(vcap_service_credentials, dict):
+                if vcap_service_credentials and isinstance(
+                        vcap_service_credentials, dict):
                     break
                 if services[key][i].get('name') == service_name:
-                    vcap_service_credentials = services[key][i].get('credentials', {})
+                    vcap_service_credentials = services[key][i].get(
+                        'credentials', {})
         if not vcap_service_credentials:
             if service_name in services.keys():
                 service = services.get(service_name)
                 if service:
-                    vcap_service_credentials = service[0].get('credentials', {})
+                    vcap_service_credentials = service[0].get(
+                        'credentials', {})
 
-        if vcap_service_credentials and isinstance(vcap_service_credentials, dict):
+        if vcap_service_credentials and isinstance(vcap_service_credentials,
+                                                   dict):
             new_vcap_creds = {}
-            if vcap_service_credentials.get('username') and vcap_service_credentials.get('password'): # cf
+            # cf
+            if vcap_service_credentials.get(
+                    'username') and vcap_service_credentials.get('password'):
                 new_vcap_creds['AUTH_TYPE'] = 'basic'
-                new_vcap_creds['USERNAME'] = vcap_service_credentials.get('username')
-                new_vcap_creds['PASSWORD'] = vcap_service_credentials.get('password')
+                new_vcap_creds['USERNAME'] = vcap_service_credentials.get(
+                    'username')
+                new_vcap_creds['PASSWORD'] = vcap_service_credentials.get(
+                    'password')
                 vcap_service_credentials = new_vcap_creds
             elif vcap_service_credentials.get('iam_apikey'):
                 new_vcap_creds['AUTH_TYPE'] = 'iam'
-                new_vcap_creds['APIKEY'] = vcap_service_credentials.get('iam_apikey')
+                new_vcap_creds['APIKEY'] = vcap_service_credentials.get(
+                    'iam_apikey')
                 vcap_service_credentials = new_vcap_creds
             elif vcap_service_credentials.get('apikey'):
                 new_vcap_creds['AUTH_TYPE'] = 'iam'
-                new_vcap_creds['APIKEY'] = vcap_service_credentials.get('apikey')
+                new_vcap_creds['APIKEY'] = vcap_service_credentials.get(
+                    'apikey')
                 vcap_service_credentials = new_vcap_creds
     return vcap_service_credentials
