@@ -1,11 +1,13 @@
 # pylint: disable=missing-docstring
 import pytest
 
-from ibm_cloud_sdk_core.authenticators import BearerTokenAuthenticator
+from ibm_cloud_sdk_core.authenticators import BearerTokenAuthenticator, Authenticator
+
 
 def test_bearer_authenticator():
     authenticator = BearerTokenAuthenticator('my_bearer_token')
     assert authenticator is not None
+    assert authenticator.authentication_type() == Authenticator.AUTHTYPE_BEARERTOKEN
     assert authenticator.bearer_token == 'my_bearer_token'
 
     authenticator.set_bearer_token('james bond')
@@ -15,6 +17,7 @@ def test_bearer_authenticator():
     authenticator.authenticate(request)
     assert request['headers']['Authorization'] == 'Bearer james bond'
 
+
 def test_bearer_validate_failed():
     with pytest.raises(ValueError) as err:
         BearerTokenAuthenticator(None)
